//V2 allows randomized, % randomized within row (but not column), and % direct across

#include "Experiments/HCUBE_BareBonesExperiment.h"

namespace HCUBE {
using namespace NEAT;



/**
 * Constructor for the experiment.
 */
BareBonesExperiment::BareBonesExperiment(string _experimentName) :
		Experiment(_experimentName),
		numNodesX((int) NEAT::Globals::getSingleton()->getParameterValue("NumNodesX-Y")), //defautl was 11 in 07 gecco papers
		numNodesY((int) NEAT::Globals::getSingleton()->getParameterValue("NumNodesX-Y")) {

	numSheets = 2;
	generateSubstrate();
}


/**
 * Function that takes a population size and creates an initial population.
 *
 * In this case each individual is a fully connected CPPN with 4 input nodes, 1 output node and a bias.
 * Every node has a sigmoid activation function. Every link has a random weight between -3 and 3.
 */
NEAT::GeneticPopulation* BareBonesExperiment::createInitialPopulation(int populationSize) {
	NEAT::GeneticPopulation* population = new NEAT::GeneticPopulation();

	vector < GeneticNodeGene > genes;

	genes.push_back(GeneticNodeGene("Bias", "NetworkSensor", 0, false));
	genes.push_back(GeneticNodeGene("X1", "NetworkSensor", 0, false));
	genes.push_back(GeneticNodeGene("Y1", "NetworkSensor", 0, false));
	genes.push_back(GeneticNodeGene("X2", "NetworkSensor", 0, false));
	genes.push_back(GeneticNodeGene("Y2", "NetworkSensor", 0, false));
	genes.push_back(GeneticNodeGene("Output", "NetworkOutputNode", 1, false, ACTIVATION_FUNCTION_SIGMOID));

	for (size_t a = 0; a < populationSize; a++) {
		shared_ptr < GeneticIndividual > individual(new GeneticIndividual(genes, true, 1.0));
		population->addIndividual(individual);
	}

	return population;
}


/**
 * Function that creates an uninitialized 'substrate'.
 *
 * The substrate is the neural network that is build using the output of the CPPN to decide on the values of the weights.
 * Note that this function does not set the weights of the substrate, that is done in the populateAndReturnSubstrate function.
 */
void BareBonesExperiment::generateSubstrate() {
	cout << "Generating substrate..." << endl;
	boost::object_pool < NEAT::NetworkNode > networkNodePool;
	boost::object_pool < NEAT::NetworkLink > networkLinkPool;

	NEAT::NetworkNode *nodes = new NEAT::NetworkNode[numNodesX * numNodesY * numSheets];
	NEAT::NetworkLink *links = new NEAT::NetworkLink[numNodesX * numNodesY * numNodesX * numNodesY * (numSheets - 1)];

	nameLookup.clear();

	cout << "Creating nodes..." << endl;

	int nodeCounter = 0;

	for (int y1 = 0; y1 < numNodesY; y1++) {
		for (int x1 = 0; x1 < numNodesX; x1++) {

			Node zeroNode(y1, x1, 0);
			Node oneNode(y1, x1, 1);

			string zeroName = toString(x1) + string("/") + toString(y1) + string("/") + toString(0);
			string oneName = toString(x1) + string("/") + toString(y1) + string("/") + toString(1);

			nameLookup[zeroNode] = zeroName;
			new (&nodes[nodeCounter]) NetworkNode(nameLookup[zeroNode]);
			nodeCounter++;

			nameLookup[oneNode] = oneName;
			new (&nodes[nodeCounter]) NetworkNode(nameLookup[oneNode]);
			nodeCounter++;
		}
	}
	cout << nodeCounter << " nodes created.\n";

	cout << "Creating links..." << endl;

	//loop through each link
	int linkCounter = 0;

	for (int y1 = 0; y1 < numNodesY; y1++) {
		for (int x1 = 0; x1 < numNodesX; x1++) {
			for (int y2 = 0; y2 < numNodesY; y2++) {
				for (int x2 = 0; x2 < numNodesX; x2++) {
					for (int z1 = 0; z1 < (numSheets - 1); z1++) {
						int z2 = z1 + 1;


						try {
							new (&links[linkCounter]) NetworkLink(
									&nodes[y1 * numNodesX * numSheets + x1 * numSheets + z1],
									&nodes[y2 * numNodesX * numSheets + x2 * numSheets + z2], 0.0);
							//LHZ - z2 will only = numSheets when recurrence is turned on - when it is, we should
							//loop the link back to the input
							linkCounter++;
						}

						catch (const char *c) {
							cout << c << endl;
							CREATE_PAUSE(string("Line Number: ") + toString(__LINE__));
						} catch (string s) {
							cout << s << endl;
							CREATE_PAUSE(string("Line Number: ") + toString(__LINE__));
						} catch (std::exception e) {
							cout << e.what() << endl;
							CREATE_PAUSE(string("Line Number: ") + toString(__LINE__));
						} catch (...) {
							cout << "AN EXCEPTION HAS OCCURED!\n";
							CREATE_PAUSE(string("Line Number: ") + toString(__LINE__));
						}

					}
				}
			}
		}
	}
	//done looping through the links

	cout << linkCounter << " links created.\n";

	cout << "Creating FastNetwork" << endl;

	substrate = NEAT::FastNetwork<double>(nodes, nodeCounter, links, linkCounter);

	cout << "done!" << endl;

	delete[] nodes;
	delete[] links;
}

/**
 * Function that takes a genetic individual and fills in the weights of its 'substrate'.
 *
 * The substrate is the neural network that is build using the output of the CPPN to decide on the values of the weights.
 * Note that this implementation can not handle multi-layered networks.
 */
NEAT::FastNetwork<double> BareBonesExperiment::populateAndReturnSubstrate(shared_ptr<const NEAT::GeneticIndividual> individual) {
	//JMC: this network is the CPPN used to produce the substrate (neural network)
	NEAT::FastNetwork<double> network = individual->spawnFastPhenotypeStack<double>();

	double cppnLinkWeightFromInputToOutputSheet = 0.0;
	double substrateLinkWeightFromInputToOutputSheet = 0.0;

	int linkCounter = 0;
	double x1Val, y1Val, x2Val, y2Val;
	for (int y1 = 0; y1 < numNodesY; y1++) {
		for (int x1 = 0; x1 < numNodesX; x1++) {
			for (int y2 = 0; y2 < numNodesY; y2++) {
				for (int x2 = 0; x2 < numNodesX; x2++) {
					//JMC: these four lines map numNodes onto a grid centered at zero, and only work for odd by odd grids
					x1Val = (x1 - numNodesX / 2) / double((numNodesX - 1) / 2);
					y1Val = (y1 - numNodesY / 2) / double((numNodesY - 1) / 2); //JMC: e.g 3 x 3 = -1 0 1, 5x5 = -2 -1 0 1 2
					x2Val = (x2 - numNodesX / 2) / double((numNodesX - 1) / 2);
					y2Val = (y2 - numNodesY / 2) / double((numNodesY - 1) / 2);

					network.reinitialize();
					network.setValue("X1", x1Val);
					network.setValue("Y1", y1Val);
					network.setValue("X2", x2Val);
					network.setValue("Y2", y2Val);
					network.setValue("Bias", 0.3);  //JMC: we have just set the inputs to the     to be x1,y1,x2,y2,bias
					network.update();               //JMC: on this line we run the CPPN network...

					Node fromInputNode(y1, x1, 0); // z=0 cause its the input layer
					Node toOutputNode(y2, x2, 1);
					Node *ptr_toNode = &toOutputNode;

					//Check whether the link indicated by the link counter is actually the link to be set.
					if ((substrate.getLink(linkCounter)) == substrate.getLink(nameLookup[fromInputNode], nameLookup[*ptr_toNode])) {
					} else {
						cout << "not the same " << linkCounter << " " << nameLookup[fromInputNode] << " " << nameLookup[*ptr_toNode] << endl;
						exit(9);
					}

					//get the cppn output and set it to the relevant link the substrate
					cppnLinkWeightFromInputToOutputSheet = network.getValue("Output");
					substrateLinkWeightFromInputToOutputSheet = zeroOrNormalize(cppnLinkWeightFromInputToOutputSheet);
					substrate.getLink(nameLookup[fromInputNode], nameLookup[*ptr_toNode])->weight = substrateLinkWeightFromInputToOutputSheet;
					linkCounter++;
				}
			}
		}
	}
	return substrate;
}

/**
 * Function that takes a weight and sets it to zero if it is within the normalization cutoff.
 * This function is used to determine when a link is removed by mutation.
 * The larger the cutoff, the larger that chance that a link is removed. (set to zero)
 */
double BareBonesExperiment::zeroOrNormalize(double mWeightToAlter) {
	//JMC: if the absolute value is greater than a threshold (here .2)...
	if (fabs(mWeightToAlter) > 0.2)
	{
		if (mWeightToAlter > 0.0)
			//JMC: ..then normalize it (0 to 3) and set it as the weight in the substrate (NN)
			return ( ((mWeightToAlter-0.2)/0.8)*3.0 );
		else
			//JMC: same thing, but if it is negative, it stays so (and is thus inhibitory)
			return ( ((mWeightToAlter+0.2)/0.8)*3.0 );
	}
	else
	{
		return (0.0);
	}
}

/**
 * Function that takes an individual and returns the result of one evaluation.
 *
 * Note that, in this implementation, we always create the substrate for this individual.
 * However, if you want to evaluate your individual multiple times, it might be better to
 * create the substrate once and only run the evaluation multiple times.
 *
 * In this bare bones implementation we use an input of 1 at every input node,
 * and the fitness is the inverse of the squared difference between the network output
 * and a checkers board pattern of ones and zeros.
 */
double BareBonesExperiment::processEvaluation(shared_ptr<NEAT::GeneticIndividual> individual) {
	populateAndReturnSubstrate (individual);
	substrate.reinitialize();
	substrate.dummyActivation();

	int xPos = 0;
	int yPos = 0;

	for (int yi = 0; yi < numNodesY; yi++){
		for (int xj = 0; xj < numNodesX; xj++) {
			xPos = xj - numNodesX / 2; //@jmc: nameLookup takes e.g. (-4,-4) and returns a string "-4/-4" which is the 'node name'
			yPos = yi - numNodesY / 2; //@jmc: the -4/-4 is simply the coordinates of the node (here -5 to 5 = 11 total, for x and y)
			Node node(yi, xj, 0);

			try {
				substrate.setValue(nameLookup[node], 1); //For this bare bones experiment the input is always 1.
			} catch (...) {
				cout << "INVALID LOCATION: " << yPos << ',' << xPos << endl;
				exit(3);
			}

		}

	}

	//Run the neural network.
	substrate.update(numSheets - 1);

	//Calculate the error
	double error = 0.0;
	double target_value;
	double output_value;

	//Go through each coordinate on the output plane and calculate the error.
	for (int y2 = 0; y2 < numNodesY; y2++)
			{
		for (int x2 = 0; x2 < numNodesX; x2++) {
			Node node(y2, x2, 1);

			//Given an input of all ones, the target is to output a checkers board pattern of ones and zeros.
			target_value = y2%2;


			output_value = substrate.getValue(nameLookup[node]); //@jmc /2 because half are neg and half are pos (cartesian grid centered on 0)
			error += ((target_value - output_value)*(target_value - output_value))/(numNodesY*numNodesX);
		}
	}

	return 1/(1+error);
}

/**
 * Function that will process a group of individuals during a certain generation.
 * Usually this group of individuals consists of a single individual,
 * but it can be manually changed in the getGroupCapacity() in Experiment.h.
 *
 * This function will be called for every individual in the population,
 * and it should take care of calculating the fitness and storing it in the individual as a reward.
 *
 * The generation parameter can be used to compare the individual in the current group
 * with other individuals in the same generation, or it can be used to asses whether we
 * are processing the last individual for this generation.
 *
 * In this bare bones implementation all we do is evaluate our individual once,
 * set this fitness as the reward for this individual, and if we are done
 * processing the last individual, we will print the champion fitness of this generation
 * and we stop the simulation if this champion got maximum fitness (solved the problem).
 */
void BareBonesExperiment::processGroup(shared_ptr<NEAT::GeneticGeneration> generation) {
	shared_ptr < NEAT::GeneticIndividual > individual = group.front();

	double fitness = processEvaluation(individual);

	individual->reward(fitness);

	if (individual == generation->getIndividual(generation->getIndividualCount() - 1)) {
		double champfit = generation->getGenerationChampion()->getFitness();
		cout << "#######GENERATION CHAMPION	" << champfit << endl;

		if (champfit >= 1) {
			cout << "PROBLEM DOMAIN SOLVED!!!\n";
			shared_ptr < NEAT::GeneticIndividual > champion = generation->getGenerationChampion();

			//print network and substrate
			printSubstrate(champion, "substrateThatSolvedProblem.txt");
			printNetworkCPPN(champion,"CPPNThatSolvedProblem.txt");

			//If the problem is solved we will stop evolution. Note that this is not always a good idea.
			SET_PARAMETER("MaxGenerations",double(generation->getGenerationNumber()+1));
		}
	}
}


/**
 * Takes an individual an prints its substrate to a file.
 */
void BareBonesExperiment::printSubstrate(shared_ptr<const NEAT::GeneticIndividual> individual, string fileName) {
	populateAndReturnSubstrate (individual);
	cout << "Printing substrate" << endl;
	ofstream substrate_file;

	substrate_file.open(fileName.c_str(), ios::trunc);

	cout << "num links: " << substrate.getLinkCount() << endl;
	substrate_file << "num links: " << substrate.getLinkCount() << endl;

	int counter = 0;
	for (int y1 = 0; y1 < numNodesY; y1++) {
		for (int x1 = 0; x1 < numNodesX; x1++) {
			for (int y2 = 0; y2 < numNodesY; y2++) {
				for (int x2 = 0; x2 < numNodesX; x2++) {
					if (counter > substrate.getLinkCount()) {
						cout << "error: you are trying to print a link that does not exist" << endl;
						exit(1);
					}

					substrate_file << "from (x1,y1) to (x2,y2): (" << x1 << "," << y1 << ") to (" << x2 << "," << y2 << "): ";
					substrate_file << substrate.getLink(counter)->weight << endl;
					counter++;
				}
			}
		}
	}
	substrate_file.close();

	return;
}

/**
 * Takes an individual an prints it CPPN to a file.
 */
void BareBonesExperiment::printNetworkCPPN(shared_ptr<const NEAT::GeneticIndividual> individual, string fileName) {
	cout << "Printing cppn network" << endl;
	ofstream network_file;
	network_file.open(fileName.c_str(), ios::trunc);

	//JMC: this creates the network CPPN associated with this individual used to produce the substrate (neural network)
	NEAT::FastNetwork<double> network = individual->spawnFastPhenotypeStack<double>();

	network_file << "num links:" << network.getLinkCount() << endl;
	network_file << "num nodes:" << network.getNodeCount() << endl;

	int numLinks = network.getLinkCount();
	int numNodes = network.getNodeCount();
	ActivationFunction activationFunction;

	//print out which node corresponds to which integer (e.g. so you can translate a fromNode of 1 to "x1"
	map<string, int> localNodeNameToIndex = *network.getNodeNameToIndex();
	for (map<string, int>::iterator iter = localNodeNameToIndex.begin(); iter != localNodeNameToIndex.end(); iter++) {
		network_file << (*iter).first << " is node number: " << (*iter).second << endl;
	}

	for (size_t a = 0; a < numLinks; a++) {

		NetworkIndexedLink<double> link = *network.getLink(a);

		network_file << link.fromNode << "->" << link.toNode << " : " << link.weight << endl;
	}
	for (size_t a = 0; a < numNodes; a++) {
		activationFunction = *network.getActivationFunction(a);
		network_file << " activation function " << a << ": ";
		if (activationFunction == ACTIVATION_FUNCTION_SIGMOID)
			network_file << "ACTIVATION_FUNCTION_SIGMOID";
		if (activationFunction == ACTIVATION_FUNCTION_SIN)
			network_file << "ACTIVATION_FUNCTION_SIN";
		if (activationFunction == ACTIVATION_FUNCTION_COS)
			network_file << "ACTIVATION_FUNCTION_COS";
		if (activationFunction == ACTIVATION_FUNCTION_GAUSSIAN)
			network_file << "ACTIVATION_FUNCTION_GAUSSIAN";
		if (activationFunction == ACTIVATION_FUNCTION_SQUARE)
			network_file << "ACTIVATION_FUNCTION_SQUARE";
		if (activationFunction == ACTIVATION_FUNCTION_ABS_ROOT)
			network_file << "ACTIVATION_FUNCTION_ABS_ROOT";
		if (activationFunction == ACTIVATION_FUNCTION_LINEAR)
			network_file << "ACTIVATION_FUNCTION_LINEAR";
		if (activationFunction == ACTIVATION_FUNCTION_ONES_COMPLIMENT)
			network_file << "ACTIVATION_FUNCTION_ONES_COMPLIMENT";
		if (activationFunction == ACTIVATION_FUNCTION_END)
			network_file << "ACTIVATION_FUNCTION_END";
		network_file << endl;
	}

	network_file.close();

	return;
}

/**
 * Creates a deep copy of the experiment.
 */
Experiment* BareBonesExperiment::clone() {
	BareBonesExperiment* experiment = new BareBonesExperiment(*this);

	return experiment;
}

}

